theory Nominal2_Lemmas
  imports Nominal2.Nominal2
begin

declare
  fresh_star_Pair[simp] fresh_star_insert[simp] fresh_Nil[simp] 
  pure_supp[simp] pure_fresh[simp]

lemma fresh_star_Nil[simp]: "{} \<sharp>* t"
  unfolding fresh_star_def by auto

lemma supp_flip[simp]:
  fixes a b :: "_ :: at" 
  shows "supp (a \<leftrightarrow> b) = (if a = b then {} else {atom a, atom b})"
  by (auto simp: flip_def supp_swap)

lemma Collect_imp: "{x. P x \<longrightarrow> Q x} = Collect (Not o P) \<union> Collect Q"
  by auto

lemma Abs_lst2_fcb2:
  fixes a b a' b' :: "atom"
  and x y :: "'b :: fs"
  and c::"'c :: fs"
  assumes e: "[[a, a']]lst. x = [[b, b']]lst. y"
  and fcb1: "a \<sharp> f a a' x c" "a' \<sharp> f a a' x c"
  and fresh: "{a, b, a', b'} \<sharp>* c"
  and perm1: "\<And>p. supp p \<sharp>* c \<Longrightarrow> p \<bullet> (f a a' x c) = f (p \<bullet> a) (p \<bullet> a') (p \<bullet> x) c"
  and perm2: "\<And>p. supp p \<sharp>* c \<Longrightarrow> p \<bullet> (f b b' y c) = f (p \<bullet> b) (p \<bullet> b') (p \<bullet> y) c"
  shows "f a a' x c = f b b' y c"
using e
apply(drule_tac Abs_lst_fcb2[where c="c" and f="\<lambda>(as::atom list) . f (hd as) (hd (tl as))"])
using fcb1 fresh perm1 perm2
apply(simp_all)
done

lemma Abs_lst2_fcb2':
  fixes a b a' b' :: "'a::at_base"
    and x y :: "'b :: fs"
    and c::"'c :: fs"
  assumes e: "[[atom a, atom a']]lst. x = [[atom b, atom b']]lst. y"
  and fcb1: "atom a \<sharp> f a a' x c" "atom a' \<sharp> f a a' x c"
  and fresh: "{atom a, atom b, atom a', atom b'} \<sharp>* c"
  and perm1: "\<And>p. supp p \<sharp>* c \<Longrightarrow> p \<bullet> (f a a' x c) = f (p \<bullet> a) (p \<bullet> a') (p \<bullet> x) c"
  and perm2: "\<And>p. supp p \<sharp>* c \<Longrightarrow> p \<bullet> (f b b' y c) = f (p \<bullet> b) (p \<bullet> b') (p \<bullet> y) c"
  shows "f a a' x c = f b b' y c"
using e
apply(drule_tac Abs_lst2_fcb2[where c="c" and f="\<lambda>a a'. f ((inv atom) a) ((inv atom) a')"])
using  fcb1 fresh perm1 perm2
apply(simp_all add: inv_f_f inj_on_def atom_eqvt)
done

lemma Abs_lst_eq_flipI:
  fixes a b :: "_ :: at" and t :: "_ :: fs"
  assumes "atom b \<sharp> t"
  shows "[[atom a]]lst. t = [[atom b]]lst. (a \<leftrightarrow> b) \<bullet> t"
  apply (auto simp: fresh_Pair)
  by (metis (no_types, lifting) assms flip_commute flip_fresh_fresh flip_triple permute_plus)

lemma Abs_lst2_eq_same_flipI:
  fixes a b :: "_ :: at"
  shows "atom b \<sharp> t \<Longrightarrow> [[atom a, atom a]]lst. t = [[atom b, atom b]]lst. (a \<leftrightarrow> b) \<bullet> t"
  apply (cases "a = b")
   apply simp
  unfolding Abs_eq_iff alphas
  apply (intro exI[of _ "(a \<leftrightarrow> b)"])
  apply (auto simp: fresh_Pair fresh_at_base supp_conv_fresh
    fresh_permute_left flip_def fresh_star_def fresh_perm swap_atom)
  done

lemma Abs_lst2_eq_flipI:
  fixes a b c d :: "'a :: at" and t :: "_ :: fs"
  assumes "atom c \<sharp> (a, b, t)" "atom d \<sharp> (a, b, t)" "a = b \<longleftrightarrow> c = d"
  shows "[[atom a, atom b]]lst. t = [[atom c, atom d]]lst. (if a = b then (a \<leftrightarrow> c) else (a \<leftrightarrow> c) + (b \<leftrightarrow> d)) \<bullet> t"
  apply (cases "a = b")
  using assms Abs_lst2_eq_same_flipI[of d t b] apply simp
  unfolding Abs_eq_iff alphas using assms
  apply (intro exI[of _ "(a \<leftrightarrow> c) + (b \<leftrightarrow> d)"])
  apply (auto simp: fresh_Pair fresh_at_base supp_conv_fresh
    fresh_permute_left flip_def fresh_star_def fresh_perm swap_atom)
  done

lemma atom_not_fresh_eq[simp]:
  assumes "\<not> atom a \<sharp> x"
  shows   "a = x"
  using assms atom_eq_iff fresh_ineq_at_base by blast

lemma fresh_set_fresh_forall:
  shows "atom y \<sharp> xs = (\<forall>x \<in> set xs. atom y \<sharp> x)"
  by (induct xs) (simp_all add: fresh_Cons)

lemma finite_fresh_set_fresh_all[simp]:
  fixes S :: "(_ :: fs) set"
  shows "finite S \<Longrightarrow> atom a \<sharp> S \<longleftrightarrow> (\<forall>x \<in> S. atom a \<sharp> x)"
  unfolding fresh_def by (simp add: supp_of_finite_sets)

lemma case_option_eqvt[eqvt]:
  "p \<bullet> case_option a b opt = case_option (p \<bullet> a) (p \<bullet> b) (p \<bullet> opt)"
  by (cases opt) auto

end
